/////////////////////////////////////////////////////////////////////////
//	"devparam.cpp"
//------------------------------------------------------------------------
//	$Id: devparam.cpp 195 2006-09-21 17:19:46Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef	HAVE_CONFIG_H
#	include	"config.h"
#	undef	HAVE_CONFIG_H
#endif
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#include	<time.h>
#include	"devtest.h"
#include	"defs.h"

#if defined(_DEBUG) && defined(WIN32)
# include	<crtdbg.h>
# define	DBGREPORT(param_)	_CrtDbgReport param_
#else
# define	DBGREPORT(param_)
#endif

DevTest::DevParam::DevParam()
{
	this->devname        = NULL;
	this->desc           = NULL;
//	this->scsi           = FALSE;
	this->devParamType   = DEVPARAMTYPE_NOTHING;
	this->inqData        = NULL;
	this->capa           = NULL;
	this->pathId         = 0;
#if		defined(WIN32)
	this->targetId       = -1;	// scan
	this->lun            = -1;	// scan
#elif	defined(LINUX)
	this->targetId       = 0;
	this->lun            = 0;
#endif
	this->setDiskinf     = FALSE;

	this->file.path      = NULL;
#if		defined(WIN32)
	this->file.handle    = INVALID_HANDLE_VALUE;
#elif	defined(LINUX)
	this->file.fd        = -1;
#endif
	this->file.blockSize = 0;
	this->file.blockNum  = 0;
	this->file.async     = FALSE;
	this->file.create    = FALSE;

#if	defined(DEVTEST_USE_MD4RNET) && defined(DEVCTL_USE_MD4RNET)
	this->net.host = NULL;
	this->net.port = 0;
#endif
	return;
}

DevTest::DevParam::~DevParam()
{
	if ( this->devname != NULL )
		delete [] this->devname;
	if ( this->desc != NULL )
		delete [] this->desc;

	if ( this->capa != NULL )
		delete [] this->capa;

#if		defined(WIN32)
	if ( this->file.handle  != INVALID_HANDLE_VALUE )
		::CloseHandle(this->file.handle);
#elif	defined(LINUX)
	if ( this->file.fd != -1 )
		::close(this->file.fd);
#endif
	if ( this->file.path != NULL )
		delete [] this->file.path;

#if	defined(DEVTEST_USE_MD4RNET) && defined(DEVCTL_USE_MD4RNET)
	if ( this->net.host != NULL )
		delete [] this->net.host;
#endif

	return;
}

int		DevTest::DevParam::SetDevName(const char * text)
{
	if ( this->devname != NULL )
	{
		delete [] this->devname;
		this->devname = NULL;
	}
	if ( text != NULL )
	{
		if ( (this->devname = new char [strlen(text)+1]) == NULL )
			return -1;
		strcpy(this->devname, text);
	}
	return 0;
}

int		DevTest::DevParam::SetDesc(const char * text)
{
	if ( this->desc != NULL )
	{
		delete [] this->desc;
			this->desc = NULL;
	}
	if ( text != NULL )
	{
		if ( (this->desc = new char [strlen(text)+1]) == NULL )
			return -1;
		strcpy(this->desc, text);
	}
	return 0;
}

const char *	DevTest::DevParam::GetDesc(unsigned max)
{
	unsigned	len;

	if ( this->desc == NULL )
		return "";
	if ( max == 0 || (len = strlen(this->desc)) <= max )
		return this->desc;
	return &this->desc[len-max];
}

int	DevTest::DevParam::SetCapacityData(DevCtl::SCSICAPACITYDATA & buf)
{
	unsigned	blockLength;
	UINT64		logicalBlockAddress;

	if ( this->capa == NULL )
	{
		if ( (this->capa = new DevCtl::SCSICAPACITYDATA16) == NULL )
			return -1;
	}
	memset(this->capa, 0, sizeof(*(this->capa)));
	blockLength = SCSI_GET32(buf.blockLength);
	SCSI_PUT32(this->capa->blockLength, blockLength);
	logicalBlockAddress = SCSI_GET32(buf.logicalBlockAddress);
	SCSI_PUT64(this->capa->logicalBlockAddress, logicalBlockAddress);

	return 0;
}

int	DevTest::DevParam::SetCapacityData(DevCtl::SCSICAPACITYDATA16 & buf)
{
	if ( this->capa == NULL )
	{
		if ( (this->capa = new DevCtl::SCSICAPACITYDATA16) == NULL )
			return -1;
	}
	*(this->capa) = buf;
	return 0;
}

void	DevTest::DevParam::ClearCapacityData(void)
{
	if ( this->capa != NULL )
	{
		delete this->capa;
		this->capa = NULL;
	}
}

int		DevTest::DevParam::ReadCapacityData(DevCtl & devCtl, unsigned * blockSize, UINT64 * blockNumber )
{
	unsigned		pathId, targetId, lun;

	if ( blockSize != NULL )
		*blockSize = 0;
	if ( blockNumber != NULL )
		*blockNumber = I64(0);
	if ( this->CheckInqData() == FALSE )
		return -1;
	pathId   = this->inqData->GetPathId();
	targetId = this->inqData->GetTargetId();
	lun      = this->inqData->GetLun();
	if ( devCtl.GetMaxCdbSize() < DevCtl::SCSI_CDB16_LEN )
	{
		DevCtl::SCSICAPACITYDATA				capa;

		if ( devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
		{	// error
			this->ClearCapacityData();
			return -1;
		} else
		{
			this->SetCapacityData(capa);
			if ( blockNumber != NULL )
				*blockNumber =  SCSI_GET32(capa.logicalBlockAddress);
			if ( blockSize != NULL )
				*blockSize = SCSI_GET32(capa.blockLength);
		}
	} else
	{
		DevCtl::SCSICAPACITYDATA16				capa;

		if ( devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
		{	// error
			this->ClearCapacityData();
			return -1;
		} else
		{
			this->SetCapacityData(capa);
			if ( blockNumber != NULL )
				*blockNumber =  SCSI_GET64(capa.logicalBlockAddress);
			if ( blockSize != NULL )
				*blockSize = SCSI_GET32(capa.blockLength);
		}
	}
	return 0;
}

int		DevTest::FileOpen(DevTest::DevParam & devParam, const char * path, INT64 blockNum, BOOL async)
{
	INT64	testsize = I64(0), filesize = I64(0);

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "DevTest::FileOpen path=%s\n", path));
	if ( devParam.CheckCapacityData() == FALSE )
		return -1;
	devParam.file.blockNum  = blockNum;
	devParam.file.blockSize = SCSI_GET32(devParam.capa->blockLength);
	if ( devParam.file.blockNum == I64(0) || devParam.file.blockSize == 0 )
		return -1;
	testsize = devParam.file.blockNum * devParam.file.blockSize;

#if		defined(WIN32)
	if ( devParam.file.handle != INVALID_HANDLE_VALUE )
		return -1;
#elif	defined(LINUX)
	if ( devParam.file.fd != -1 )
		return -1;
#endif

	{
		DevTest::FileStatInfo	finf;
#if		defined(WIN32)
		DWORD					dwCreate = CREATE_NEW;
#elif	defined(LINUX)
		int						flags = O_CREAT | O_EXCL;
#endif
		int						ret;

		if ( DevTest::FileStat(path, finf) == 0 )
		{
			DevTest::fprintf(stderr, "File already exist! (%s)\n", path);
			if ( finf.attr_dir )
				return -1;
			if ( this->opt.noprompt == FALSE )
			{
				fprintf(stderr, "[D:Delete [O:Overwrite [A:Abort - Select ? ");
				ret = DevTest::SelectConsole("DOA");
			} else
				ret = 'O';	// overwrite
			switch  (ret )
			{
			case 'D':	// Delete
#if		defined(WIN32)
				::DeleteFile(path);
				dwCreate = CREATE_NEW;
#elif	defined(LINUX)
				remove(path);
#endif
				break;
			case 'O':	// Overwrite
#if		defined(WIN32)
				dwCreate = OPEN_EXISTING;
#elif	defined(LINUX)
				flags = 0;
#endif
				filesize = finf.filesize;
				break;

			case 'A':	// Abort
				return -1;
			}
		}
#if		defined(WIN32)
		devParam.file.handle = ::CreateFile(
		    path
		  , GENERIC_READ | GENERIC_WRITE
		  , FILE_SHARE_READ
		  , NULL
		  , dwCreate
		  , FILE_ATTRIBUTE_NORMAL | FILE_FLAG_WRITE_THROUGH | FILE_FLAG_NO_BUFFERING | (async ? FILE_FLAG_OVERLAPPED : 0)
		  , NULL );
		if ( devParam.file.handle == INVALID_HANDLE_VALUE )
			return -1;
		if ( dwCreate == CREATE_NEW )
			devParam.file.create = TRUE;
		else
			devParam.file.create = FALSE;
#elif	defined(LINUX)
		if ( (devParam.file.fd = open(path, flags | O_RDWR | O_SYNC)) < 0 )
		{
			devParam.file.fd = -1;
			return -1;
		}
		if ( (flags & O_CREAT) && (flags & O_EXCL) )
			devParam.file.create = TRUE;
		else
			devParam.file.create = FALSE;
#endif
	}

	devParam.file.async = async;
	{
		if ( devParam.file.path != NULL )
			delete [] devParam.file.path;
		if ( (devParam.file.path = new char [strlen(path) + 1]) == NULL )
		{
			this->FileClose(devParam);
			return -1;
		}
		strcpy(devParam.file.path, path);
	}

	{
		BYTE	*buffer = NULL;
		INT64	addr = filesize / devParam.file.blockSize;
		time_t	tmNow, tmLast;

		fprintf(stderr, "addr %"FI64"u\n", addr);
		if ( (buffer = new BYTE [devParam.file.blockSize]) == NULL )
			return -1;
		memset(buffer, 0, devParam.file.blockSize);
		time(&tmLast);
		do
		{
			time(&tmNow);
			if ( difftime(tmNow, tmLast) > 1 )
			{
				DevTest::fprintf(stderr, "Create %16llu/%-16llu (%3u%%)\r"
				  , addr * devParam.file.blockSize
				  , devParam.file.blockNum * devParam.file.blockSize
				  , (unsigned)( I64(100) * addr / devParam.file.blockNum) );
				tmLast = tmNow;
			}
			addr += (INT64)(TESTFILE_CREATESIZE / devParam.file.blockSize);
			if ( addr > devParam.file.blockNum - I64(1) )
				addr = devParam.file.blockNum - I64(1);
			devParam.FileWrite( addr, 1, devParam.file.blockSize, buffer);
		} while ( addr < devParam.file.blockNum - I64(1) );
		DevTest::fprintf(stderr, "\n");
		delete [] buffer;
	}

	return 0;
}

int		DevTest::FileClose(DevTest::DevParam & devParam)
{
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "DevTest::FileClose\n"));
#if		defined(WIN32)
	if ( devParam.file.handle != INVALID_HANDLE_VALUE )
	{
		::CloseHandle(devParam.file.handle);
		devParam.file.handle = INVALID_HANDLE_VALUE;
		if ( devParam.file.path != NULL && devParam.file.create )
			::DeleteFile(devParam.file.path);
	}
#elif	defined(LINUX)
	if ( devParam.file.fd != -1 )
	{
		::close(devParam.file.fd);
		devParam.file.fd = -1;
		if ( devParam.file.path != NULL && devParam.file.create )
			::remove(devParam.file.path);
	}
#endif
	if ( devParam.file.path != NULL )
	{
		delete [] devParam.file.path;
		devParam.file.path = NULL;
	}

	return 0;
}


int		DevTest::DevParam::FileRead(INT64 addr, unsigned count, unsigned size, void * buffer)
{
#if		defined(WIN32)
	if ( this->file.handle != INVALID_HANDLE_VALUE )
	{
		LARGE_INTEGER	pos;
		DWORD			dwResult;

		if ( addr + (INT64)count > this->file.blockNum )
			return -1;
		pos.QuadPart = addr * (INT64) this->file.blockSize;
		if ( this->file.async )
		{
			OVERLAPPED		ol;

			memset(&ol, 0, sizeof(ol));
			ol.Offset     = pos.LowPart;
			ol.OffsetHigh = pos.HighPart;
			if ( ::ReadFile(this->file.handle, buffer, size, &dwResult, &ol) == FALSE )
			{
				if ( ::GetLastError() != ERROR_IO_PENDING )
					return -1;
			}
			if ( ::GetOverlappedResult(this->file.handle, &ol, &dwResult, TRUE) == 0 )
				return -1;
			dwResult = size;
		} else
		{
			if ( pos.LowPart = ::SetFilePointer(this->file.handle, pos.LowPart, &pos.HighPart, FILE_BEGIN) == INVALID_SET_FILE_POINTER
			  && ::GetLastError() != NO_ERROR )
				return -1;
			if ( ::ReadFile(this->file.handle, buffer, size, &dwResult, NULL) == FALSE )
				return -1;
		}
		return (int)dwResult;
	}
#elif	defined(LINUX)
	if ( this->file.fd != -1 )
	{
		off_t			pos;
		ssize_t			res;

		if ( addr + (INT64)count > this->file.blockNum )
			return -1;
		pos = addr * (INT64) this->file.blockSize;
		if ( lseek(this->file.fd, pos, SEEK_SET) < 0 )
			return -1;
		if ( (res = read(this->file.fd, buffer, size)) != size )
			return -1;
		return (int) res;
	}
#endif

	return -1;
}

int		DevTest::DevParam::FileWrite(INT64 addr, unsigned count, unsigned size, const void * buffer)
{
#if		defined(WIN32)
	if ( this->file.handle != INVALID_HANDLE_VALUE )
	{
		LARGE_INTEGER	pos;
		DWORD			dwResult;

		if ( addr + (INT64)count > this->file.blockNum )
			return -1;
		pos.QuadPart = addr * (INT64) this->file.blockSize;
		if ( this->file.async )
		{
			OVERLAPPED		ol;

			memset(&ol, 0, sizeof(ol));
			ol.Offset     = pos.LowPart;
			ol.OffsetHigh = pos.HighPart;
			if ( ::WriteFile(this->file.handle, buffer, size, &dwResult, &ol) == FALSE )
			{
				if ( ::GetLastError() != ERROR_IO_PENDING )
					return -1;
			}
			if ( ::GetOverlappedResult(this->file.handle, &ol, &dwResult, TRUE) == 0 )
				return -1;
			dwResult = size;
		} else
		{
			if ( pos.LowPart = ::SetFilePointer(this->file.handle, pos.LowPart, &pos.HighPart, FILE_BEGIN) == INVALID_SET_FILE_POINTER
			  && ::GetLastError() != NO_ERROR )
				return -1;
			if ( ::WriteFile(this->file.handle, buffer, size, &dwResult, NULL) == FALSE )
				return -1;
		}
		return (int)dwResult;
	}
#elif	defined(LINUX)
	if ( this->file.fd != -1 )
	{
		off_t			pos;
		ssize_t			res;

		if ( addr + (INT64)count > this->file.blockNum )
			return -1;
		pos = addr * (INT64) this->file.blockSize;
		if ( lseek(this->file.fd, pos, SEEK_SET) < 0 )
			return -1;
		if ( (res = write(this->file.fd, buffer, size)) != size )
			return -1;
		return (int) res;
	}
#endif

	return -1;
}

#if	defined(DEVTEST_USE_MD4RNET) && defined(DEVCTL_USE_MD4RNET)
int			DevTest::DevParam::SetNet(const char * host, int port, int pathId, int targetId, int lun)
{
	if ( host == NULL )
		return -1;
	if ( this->net.host != NULL )
		delete [] this->net.host;
	if ( (this->net.host = new char [strlen(host)+1]) == NULL )
		return -1;
	strcpy(this->net.host, host);
	this->net.port     = port;
	this->pathId       = pathId;
	this->targetId     = targetId;
	this->lun          = lun;
	this->devParamType = DEVPARAMTYPE_NET;

	return 0;
}
#endif

