/////////////////////////////////////////////////////////////////////////
//	"devtest_udf.cpp"
//------------------------------------------------------------------------
//	$Id: devtest_udf.cpp 187 2006-09-21 04:42:00Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef	HAVE_CONFIG_H
#	include	"config.h"
#	undef	HAVE_CONFIG_H
#endif
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	"devtest.h"
#include	"defs.h"

#if	defined(DEVTEST_USE_UDFLIB)

#include	"udflib.h"

static char *	FmtUdfTime(char * buf, UdfLib::timestamp & ts)
{
	sprintf(buf, "[%2u:%4u] %04u/%02u/%02u %02u:%02u:%02u (%02u.%02u.%02u)"
	  , (UdfLib::U_endian16(ts.typeAndTimeZone)>>12) & 15
	  , UdfLib::U_endian16(ts.typeAndTimeZone) & (4095)
	  , UdfLib::U_endian16(ts.year)
	  , ts.month
	  , ts.day
	  , ts.hour
	  , ts.minute
	  , ts.second
	  , ts.centiseconds
	  , ts.hundredMicroseconds
	  , ts.microseconds
	);
	return buf;
}

#if	defined(LINUX)
#define	CP_OEMCP	(0)
// Windows API Like
static int	WideCharToMultiByte(unsigned codepage, DWORD flags, const wchar_t * wbuf, int wlen, char * abuf, int abyte, const char *defchar, BOOL * useDefChar)
{
	int				wch, wpos, alen = 0;

	if ( wbuf == NULL )
		return 0;	// error
	for ( wpos = 0; wpos < wlen; wpos++ )
	{
		if ( abuf != NULL && alen >= abyte )
			break;
		wch = wbuf[wpos];
		if ( wch < 0x100 )
		{
			if ( abuf != NULL && alen < abyte )
				abuf[alen] = (char)wch;
			alen++;
		} else
		{
			if ( abuf != NULL && alen + 1 < abyte )
			{
				abuf[alen  ] = '@';
				abuf[alen+1] = '@';
			}
			alen += 2;
		}
	}
	return alen;
}
#endif

static char *	UnCompUnicode(char * buffer, unsigned bytes, const BYTE * src)
{
	int		ret;
	wchar_t	*wbuf;

	buffer[0] = '\0';
	if ( (wbuf = new wchar_t [bytes]) != NULL )
	{
		ret = UdfLib::UncompressUnicode(bytes, (UdfLib::Uint8 *)src, (UdfLib::unicode_t *)wbuf);
		if ( ret > 0 )
		{
			ret = ::WideCharToMultiByte(CP_OEMCP, 0, wbuf, ret, buffer, bytes, NULL, NULL);
			if ( ret < (int)bytes )
				buffer[ret] = '\0';
		} else
			memcpy(buffer, src, bytes);
		delete [] wbuf;
	}
	return buffer;
}

int		DevTest::PrintDisc(DevTest::DevParam & devParam)
{
	unsigned					pathId, targetId, lun;
	BYTE						*buffer = NULL;
	INT64						blkNum;
	unsigned					blkLen;
	unsigned					blockSize;
	UdfLib::VRSInfo				vrsInfo;
	BOOL						flag_disc = FALSE, flag_bea = FALSE;

	if ( devParam.CheckInqData() == FALSE || devParam.CheckCapacityData() == FALSE )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	blkNum = devParam.GetCapacityBlockAddress();
	blkLen = devParam.GetCapacityBlockLength();
	blockSize = blkLen;
	this->udflib.OpenUdf(this->devCtl, *(devParam.GetInqData()), *(devParam.GetCapacityData()), I64(0), blkNum);
	if ( this->udflib.CheckVRS(vrsInfo) == 0 )
	{
		UdfLib::VRSInfo::Vrs	*vrs;

		// "  IBM HARD-DISK FORMAT    (########) ########## sectors   ####################B
		fprintf(stdout, "  CD/DVD DISC FORMAT      (%8u) %10"FI64"u sectors   %20,I64uB\n"
		  , blockSize
		  , (INT64)blkNum
		  , (INT64)blkNum * (INT64)blockSize
		);
		if ( this->opt.printDetail == FALSE )
			return 0;
		flag_disc = TRUE;
		for ( vrs = vrsInfo.vrsTop; vrs != NULL ; vrs = vrs->next )
		{
			fprintf(stdout, "    Found ID:%.5s (sector %3u) ver.%02XH :", vrs->vrs.sid, vrs->loc, vrs->vrs.version);
			switch ( vrs->type )
			{
				case UdfLib::SID_BEA01:
					flag_bea = TRUE;
					break;
				case UdfLib::SID_BOOT2:
					break;
				case UdfLib::SID_CD001:
					switch ( vrs->vrs.type )
					{
						case 0x00: fprintf(stdout, "ISO9660 Boot Record"                     ); break;
						case 0x01: fprintf(stdout, "ISO9660 Primary Volume Descriptor"       ); break;
						case 0x02: fprintf(stdout, "ISO9660 Supplementary Volume Descriptor" ); break;
						case 0x03: fprintf(stdout, "ISO9660 Volume Partition Descriptor"     ); break;
						case 0xFF: fprintf(stdout, "ISO9660 Volume Descriptor Set Terminator"); break;
						default  : fprintf(stdout, "ISO9660 %02H", vrs->type                 ); break;
					}
					break;
				case UdfLib::SID_CDW02:
					break;
				case UdfLib::SID_NSR02:
					break;
				case UdfLib::SID_NSR03:
					break;
				case UdfLib::SID_TEA01:
					break;
			}
			fprintf(stdout, "\n");
		}
	}
	if ( flag_bea )
	{
		UdfLib::ADVPInfo	avdpInfo;
		UdfLib::VdsInfo		vdsInfo;

		memset(&avdpInfo, 0, sizeof(avdpInfo));
		if ( this->udflib.GetAVDP(avdpInfo) == 0 && (avdpInfo.flag & 3) != 0 )
		{
			fprintf(stdout, "  Anchor Volume Descriptor Pointer : \n");
			if ( (avdpInfo.flag & 1) != 0 )
			{
				fprintf(stdout, "        Front %10u (loc:%u len:%u) OK\n" 
				  , (unsigned)avdpInfo.p[0].loc
				  , this->udflib.U_endian32(avdpInfo.p[0].avdp.mainVDSAdr.location)
				  , this->udflib.U_endian32(avdpInfo.p[0].avdp.mainVDSAdr.length  )
				);
			}
			if ( (avdpInfo.flag & 2) != 0 )
			{
				fprintf(stdout, "        Back  %10u (loc:%u len:%u) OK\n" 
				  , (unsigned)avdpInfo.p[1].loc
				  , this->udflib.U_endian32(avdpInfo.p[1].avdp.mainVDSAdr.location)
				  , this->udflib.U_endian32(avdpInfo.p[1].avdp.mainVDSAdr.length  )
				);
			}
			if ( avdpInfo.flag & 1 )
			{
				if ( this->udflib.GetVDS(vdsInfo, avdpInfo.p[0].avdp) == 0 )
				{
					UdfLib::VolDesInfo	*volDesInfo;
					UdfLib::VolDes		*volDes;

					fprintf(stdout, "  Volume Descriptor Sequence\n");
					volDesInfo = &vdsInfo.mVolDes;
					for ( volDes = volDesInfo->desTop; volDes != NULL; volDes = volDes->next )
					{
						char	text[256], t1[256], t2[256];

						sprintf(text, "    loc:%u type:%u : ", volDes->loc, volDes->type);
						switch ( volDes->type )
						{
							case UdfLib::VOLDES_PVD:
							{
								UdfLib::PrimaryVolDes	*pvd;

								pvd = &volDes->des->pvd;
								fprintf(stdout, "%s Primary Volume Descriptor\n", text);
								PutPos(8, "Volume Descriptor Sequence Number", COLUMN1+6, Fmt(t1,"%4u", this->udflib.U_endian32(pvd->volDescSeqNum)) ,0);
								PutPos(8, "PVD Number"                  , COLUMN1, Fmt(t1,"%10u", this->udflib.U_endian32(pvd->primVolDesNum)) ,0);
								PutPos(8, "Volume ID"                   , COLUMN1, Fmt(t1,"%.32s" , UnCompUnicode(t2,32,pvd->volID))    , 0);
								PutPos(8, "Volume Sequence Number"      , COLUMN1, Fmt(t1,"%10u", this->udflib.U_endian16(pvd->volSeqNum          )) ,0);
								PutPos(8, "Max Volume Sequence Number"  , COLUMN1, Fmt(t1,"%10u", this->udflib.U_endian16(pvd->maxVolSeqNum       )) ,0);
								PutPos(8, "Interchange Level"           , COLUMN1, Fmt(t1,"%10u", this->udflib.U_endian16(pvd->interchangeLevel   )) ,0);
								PutPos(8, "Max Interchange Level"       , COLUMN1, Fmt(t1,"%10u", this->udflib.U_endian16(pvd->maxInterchangeLevel)) ,0);
								PutPos(8, "Charactor Set List"          , COLUMN1, Fmt(t1,"%10u", this->udflib.U_endian32(pvd->charSetList        )) ,0);
								PutPos(8, "Max Charactor Set List"      , COLUMN1, Fmt(t1,"%10u", this->udflib.U_endian32(pvd->maxCharSetList     )) ,0);
								PutPos(8, "Volume Set ID"               , COLUMN1, Fmt(t1,"%.128s", UnCompUnicode(t2,128,pvd->volSetID)) , 0);
								PutPos(8, "Descriptor CharSet"          , COLUMN1, Fmt(t1,"%.02XH: %.63s", pvd->desCharSet.charSetType        , pvd->desCharSet.charSetInfo        ), 0);
								PutPos(8, "Explanatory CharSet"         , COLUMN1, Fmt(t1,"%.02XH: %.63s", pvd->explanatoryCharSet.charSetType, pvd->explanatoryCharSet.charSetInfo), 0);
								if ( pvd->volAbstract.location || pvd->volAbstract.length )
								{
									PutPos(8, "Volume Abstract"             
									  ,COLUMN1, Fmt(t1,"loc:%u len:%u"
									  ,this->udflib.U_endian32(pvd->volAbstract.location)
									  ,this->udflib.U_endian32(pvd->volAbstract.length))
									, 0);
								}
								if ( pvd->volCopyrightNotice.location || pvd->volCopyrightNotice.length )
								{
									PutPos(8, "Volume Copyright Notice"     
									  ,COLUMN1, Fmt(t1,"loc:%u len:%u"
									  ,this->udflib.U_endian32(pvd->volCopyrightNotice.location)
									  ,this->udflib.U_endian32(pvd->volCopyrightNotice.length  ))
									,0);
								}
								PutPos(8, "Application ID"              , COLUMN1, Fmt(t1,"%02XH: %.31s" , pvd->applicationID.flags, pvd->applicationID.id), 0);
								PutPos(8, "Recording Date/Time"         , COLUMN1-10, FmtUdfTime(t1, pvd->recordingTime), 0);
								PutPos(8, "Implementation ID"           , COLUMN1, Fmt(t1,"%02XH: %.23s" , pvd->implementationID.flags, pvd->implementationID.id), 0);
								PutPos(8, ""                            , COLUMN1, Fmt(t1,"OS Class:%02XH  OS ID:%02XH" , pvd->implementationID.OSClass, pvd->implementationID.OSIdentifier), 0);
								if ( pvd->predecessorVDSLoc )
									PutPos(8, "Predecessor VDS location", COLUMN1, Fmt(t1,"%10u", this->udflib.U_endian32(pvd->predecessorVDSLoc)) ,0);
								break;
							}
							case UdfLib::VOLDES_LVD:
							{
								UdfLib::LogVolDesc	*lvd;

								lvd = &volDes->des->lvd;
								fprintf(stdout,"%s Logical Volume Descriptor\n", text);
								PutPos(8, "Volume Descriptor Sequence Number", COLUMN1+6, Fmt(t1,"%4u", this->udflib.U_endian32(lvd->volDescSeqNum)) ,0);
								PutPos(8, "Descriptor CharSet"          , COLUMN1, Fmt(t1,"%.02XH: %.63s", lvd->desCharSet.charSetType, lvd->desCharSet.charSetInfo), 0);
								PutPos(8, "Logical Volume ID"  , COLUMN1, Fmt(t1,"%.128s" , UnCompUnicode(t2,128,lvd->logVolID))   , 0);
								PutPos(8, "Logical Block Size" , COLUMN1, Fmt(t1,"%10u"   , this->udflib.U_endian32(lvd->logBlkSize)), 0);
								PutPos(8, "Domain ID"          , COLUMN1, Fmt(t1,"%02XH: %.23s" , lvd->domainID.flags, lvd->domainID.id), 0);
								PutPos(8, "UDF Revision"       , COLUMN1+6, Fmt(t1,"%04XH Domain flags:%02XH" , this->udflib.U_endian16(lvd->domainID.UDFRevision), lvd->domainID.domainFlags), 0);
							//	PutPos(8, "Logical Volume Contents Use", COLUMN1, Fmt(t1,"%.16s", lvd->logVolUse) , 0);
								PutPos(8, "Map Table Length"   , COLUMN1, Fmt(t1,"%10u"   , lvd->mapTabLen) , 0);
								PutPos(8, "Number of Partition Maps", COLUMN1, Fmt(t1,"%10u"   , lvd->numPartMaps) , 0);
								PutPos(8, "Implementation ID"           , COLUMN1, Fmt(t1,"%02XH: %.23s" , lvd->implementationID.flags, lvd->implementationID.id), 0);
								PutPos(8, ""                            , COLUMN1, Fmt(t1,"OS Class:%02XH  OS ID:%02XH" , lvd->implementationID.OSClass, lvd->implementationID.OSIdentifier), 0);
							//	PutPos(8, "Implementation Use" , COLUMN1, Fmt(t1,"%.128s", lvd->implementationUse) , 0);
								if ( lvd->integritySeqExtent.location || lvd->integritySeqExtent.length )
								{
									PutPos(8, "Integrity Sequence Extent"     
									  ,COLUMN1, Fmt(t1,"loc:%u len:%u"
									  ,this->udflib.U_endian32(lvd->integritySeqExtent.location)
									  ,this->udflib.U_endian32(lvd->integritySeqExtent.length  ))
									,0);
								}
								break;
							}
							case UdfLib::VOLDES_IUVD:
							{
								UdfLib::ImpUseVolDesc	*iuvd;
								UdfLib::LVInformation	*lvi;

								iuvd = (UdfLib::ImpUseVolDesc *)&volDes->des->iuvd;
								fprintf(stdout,"%s Implementation Use Volume Descriptor\n", text);
								PutPos(8, "Volume Descriptor Sequence Number", COLUMN1+6, Fmt(t1,"%4u", this->udflib.U_endian32(iuvd->volDescSeqNum)) ,0);
								PutPos(8, "Implementation ID"           , COLUMN1, Fmt(t1,"%02XH: %.23s" , iuvd->implementationID.flags, iuvd->implementationID.id), 0);
								PutPos(8, "UDF Revision"       , COLUMN1+6, Fmt(t1,"%04XH" , this->udflib.U_endian16(iuvd->implementationID.UDFRevision)), 0);
								PutPos(8, ""                            , COLUMN1, Fmt(t1,"OS Class:%02XH  OS ID:%02XH" , iuvd->implementationID.OSClass, iuvd->implementationID.OSIdentifier), 0);
								lvi = (UdfLib::LVInformation *)iuvd->implementationUse;
								PutPos(8, "LVICharset"          , COLUMN1, Fmt(t1,"%.02XH: %.63s", lvi->LVICharset.charSetType, lvi->LVICharset.charSetInfo), 0);
								PutPos(8, "Logical Volume ID"  , COLUMN1, Fmt(t1,"%.128s" , UnCompUnicode(t2,128,lvi->logVolId))   , 0);
								if ( lvi->LVInfo1[0] || lvi->LVInfo2[0] || lvi->LVInfo3[0] )
								{
									PutPos(8, "LV Info 1"  , COLUMN1, Fmt(t1,"%.36s" , UnCompUnicode(t2,36,lvi->LVInfo1)), 0);
									PutPos(8, "LV Info 2"  , COLUMN1, Fmt(t1,"%.36s" , UnCompUnicode(t2,36,lvi->LVInfo1)), 0);
									PutPos(8, "LV Info 3"  , COLUMN1, Fmt(t1,"%.36s" , UnCompUnicode(t2,36,lvi->LVInfo1)), 0);
								}
								PutPos(8, "Implementation ID"           , COLUMN1, Fmt(t1,"%02XH: %.23s" , lvi->implementationID.flags, lvi->implementationID.id), 0);
								PutPos(8, ""                            , COLUMN1, Fmt(t1,"OS Class:%02XH  OS ID:%02XH" , lvi->implementationID.OSClass, lvi->implementationID.OSIdentifier), 0);
								break;
							}
							case UdfLib::VOLDES_USD:
							{
								UdfLib::UnallocSpDesHead	*usd;
								unsigned					i;

								usd = (UdfLib::UnallocSpDesHead *)&volDes->des->usd;
								fprintf(stdout,"%s Unallocated Space Entry\n", text);
								PutPos(8, "Volume Descriptor Sequence Number", COLUMN1+6, Fmt(t1,"%4u", this->udflib.U_endian32(usd->volDescSeqNum)) ,0);
								PutPos(8, "Number of Allocation Descriptor"  , COLUMN1+6, Fmt(t1,"%4u", this->udflib.U_endian32(usd->numAllocationDes)), 0);
								for ( i = 0; i < this->udflib.U_endian32(usd->numAllocationDes); i++ )
								{
									PutPos(8, "Allocation Descriptor"
									  ,COLUMN1, Fmt(t1,"No.%3u loc:%u len:%u"
									  ,i + 1
									  ,this->udflib.U_endian32(usd->allocationDes[i].location)
									  ,this->udflib.U_endian32(usd->allocationDes[i].length  ))
									,0);
								}
								break;
							}
							case UdfLib::VOLDES_PD :
							{
								UdfLib::PartDesc	*pd;

								pd = (UdfLib::PartDesc *)&volDes->des->pd;
								fprintf(stdout,"%s Partition Descriptor\n", text);
								PutPos(8, "Volume Descriptor Sequence Number", COLUMN1+6, Fmt(t1,"%4u", this->udflib.U_endian32(pd->volDescSeqNum)) ,0);
								PutPos(8, "Partition Flags", COLUMN1+6, Fmt(t1,"%04XH", this->udflib.U_endian16(pd->partFlags)) ,0);
								PutPos(8, "Partition Number", COLUMN1+6, Fmt(t1,"%04u", this->udflib.U_endian16(pd->partNumber)) ,0);
								PutPos(8, "Partition Contents"              , COLUMN1, Fmt(t1,"%02XH: %.31s" , pd->partContents.flags, pd->partContents.id), 0);
								PutPos(8, "Access Type", COLUMN1+2, Fmt(t1,"%08XH", this->udflib.U_endian32(pd->accessType)) ,0);
								PutPos(8, "Partition Starting"     
								  ,COLUMN1, Fmt(t1,"loc:%u len:%u"
								  ,this->udflib.U_endian32(pd->partStartingLoc)
								  ,this->udflib.U_endian32(pd->partLength     ) )
								,0);
								PutPos(8, "Implementation ID"           , COLUMN1, Fmt(t1,"%02XH: %.23s" , pd->implementationID.flags, pd->implementationID.id), 0);
								PutPos(8, ""                            , COLUMN1, Fmt(t1,"OS Class:%02XH  OS ID:%02XH" , pd->implementationID.OSClass, pd->implementationID.OSIdentifier), 0);
								break;
							}
							default:
								fprintf(stdout,"%s Unknown Descriptor\n", text);
								break;
						}
					}
				}
			}
		}
	} // end of flag_bea
	return flag_disc ? 0 : -1;
}

#endif

