/////////////////////////////////////////////////////////////////////////
//	"ext2lib_lib.cpp"
//------------------------------------------------------------------------
//	$Id: ext2lib.cpp 176 2006-05-26 12:05:14Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef	HAVE_CONFIG_H
#	include	"config.h"
#	undef	HAVE_CONFIG_H
#endif
#include	<stdlib.h>
#include	<string.h>
#include	<stdio.h>
#include	"ext2lib.h"

Ext2Lib::Ext2Lib()
{
	this->devCtl     = NULL;
	this->inqData    = NULL;
	this->capa       = NULL;
	this->sec_off    = 0;
	this->sec_num    = 0;
	this->iobuf_size = 0;
	this->iobuf      = NULL;
	this->buffer     = NULL;
}

Ext2Lib::~Ext2Lib()
{
	if ( this->iobuf != NULL )
		delete [] this->iobuf;
	if ( this->buffer != NULL )
		delete [] this->buffer;
}

int		Ext2Lib::OpenExt2(DevCtl & devCtl, DevCtl::ScsiInqData & inq, DevCtl::SCSICAPACITYDATA16 & capa, INT64 sec_off, INT64 sec_num)
{
	unsigned			blkLen;

	this->devCtl  = &devCtl;
	this->inqData = &inq;
	this->capa    = &capa;
	this->sec_off = sec_off;
	this->sec_num = sec_num;
	if ( this->sec_off + this->sec_num > this->devCtl->ScsiGet64(this->capa->logicalBlockAddress) )
		return -1;	// error
	blkLen = this->GetBlkSize();
	this->iobuf_size = BLOCK_SIZE > blkLen
	  ? ((BLOCK_SIZE + blkLen    -1) / blkLen    ) * blkLen
	  : ((blkLen     + BLOCK_SIZE-1) / BLOCK_SIZE) * BLOCK_SIZE;
	this->iobuf_size *= 2;
	if ( this->iobuf != NULL )
		delete [] this->iobuf;
	if ( (this->iobuf = new BYTE [this->iobuf_size]) == NULL )
		return -1;
	if ( this->buffer != NULL )
		delete [] this->buffer;

	if ( (this->buffer = new BYTE [BLOCK_SIZE]) == NULL )
		return -1;

	return 0;
}

int		Ext2Lib::CheckSuperBlock(Ext2Lib::ext2_super_block & superblock)
{
	ext2_super_block	*sb;
	INT64				addr;
	int					i;

	if ( this->devCtl == NULL || this->capa == NULL 
	  || this->iobuf == NULL || this->buffer == NULL )
		return -1;
	addr = 1;
	for ( i = 0; i < CHECK_SUPERBLOCK_NUM; i++ )
	{
		if ( this->ReadBlock( 1, 1, this->buffer) < 0 )
			continue;
		sb = (ext2_super_block *) this->buffer;
		if ( sb->s_magic != EXT2_SUPER_MAGIC )
			continue;
		memcpy( &superblock, this->buffer, sizeof(superblock));
		return 0;	// OK
	}
	return -1;	// error
}


TCHAR	*	Ext2Lib::GetOsName(unsigned os, TCHAR * buffer)
{
	switch ( os )
	{
		case OS_LINUX  : _tcscpy(buffer, TEXT("Linux"  )); break;
		case OS_HURD   : _tcscpy(buffer, TEXT("Hurd"   )); break;
		case OS_MASIX  : _tcscpy(buffer, TEXT("Masix"  )); break;
		case OS_FREEBSD: _tcscpy(buffer, TEXT("FreeBSD")); break;
		case OS_LITES  : _tcscpy(buffer, TEXT("Lites"  )); break;
		default        : return NULL;
	}
	return buffer;
}

int		Ext2Lib::ReadBlock(INT64 blkAddr, unsigned blkCount, void * buffer)
{
	return this->Read(blkAddr * BLOCK_SIZE, blkCount * BLOCK_SIZE, buffer);
}

int		Ext2Lib::Read(INT64 addr, unsigned size, void * buffer)
{
	INT64		sec_addr;
	unsigned	sec_count, sec_off, sec_size, blkLen;

	blkLen    = this->GetBlkSize();
	sec_addr  = addr / blkLen;
	sec_off   = (unsigned)(addr % (INT64)blkLen);
	sec_count = (size + blkLen - 1) / blkLen;
	sec_size  = sec_count * blkLen;
	if ( sec_size > this->iobuf_size )
		return -1;
	if ( this->ReadSector(sec_addr, sec_count, sec_size, this->iobuf) < 0 )
		return -1;
	memcpy( buffer, &this->iobuf[sec_off], size);
	return 0;
}

int		Ext2Lib::ReadSector(INT64 addr, unsigned count, unsigned size, void * buffer)
{
	if ( this->devCtl == NULL || this->inqData == NULL )
		return -1;
//	fprintf(stderr, "Read %llu\n", (INT64) this->sec_off + addr);	// DEBUG
	if ( this->devCtl->ScsiRead(
	  this->inqData->GetPathId(),
	  this->inqData->GetTargetId(),
	  this->inqData->GetLun(),
	  this->sec_off + addr,
	  count,
	  size,
	  buffer ) < (int)size )
		return -1;
	return 0;
}

int		Ext2Lib::GetBlkSize(void)
{
	if ( this->devCtl == NULL || this->capa == NULL )
		return -1;
	return (int)this->devCtl->ScsiGet32(capa->blockLength);
}

INT64	Ext2Lib::GetBlkNum(void)
{
	if ( this->devCtl == NULL || this->capa == NULL )
		return -1;
	return this->sec_num;
//	return this->devCtl->ScsiGetLBA(capa->logicalBlockAddress) - (this->sec_off + this->sec_max);
}

