/*************************************************************************
*	"crc16xm.c"
*-------------------------------------------------------------------------
*	$Id: crc16xm.c 94 2006-09-20 08:51:21Z taro $
*************************************************************************/

#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif
#include	"crc16.h"
#include	<stdlib.h>
#include	<limits.h>

#define	XM_CRCPOLY	0x1021u		/* x^{16}+x^{12}+x^5+1 */

static int		crc16xm_table_init = 0;
unsigned int	crc16xm_table[UCHAR_MAX + 1];

#ifndef	BYTE
#	define	BYTE	unsigned char
#endif

static void crc16xm_maketable(void)
{
	unsigned int		i, j, r;

	for ( i = 0; i <= UCHAR_MAX; i++)
	{
		r = i << (16 - CHAR_BIT);
		for (j = 0; j < CHAR_BIT; j++)
			if ( r & 0x8000u)
				r = (r << 1) ^ XM_CRCPOLY;
			else
				r <<= 1;
		crc16xm_table[i] = r & 0xFFFFu;
	}
	return;
}

void		crc16xm_init(Crc16work * crcwork)
{
	if ( crcwork == NULL )
		return;
	if ( crc16xm_table_init == 0 )
	{
		crc16xm_maketable();
		crc16xm_table_init = 1;
	}
	crcwork->r = 0xFFFFu;
	return;
}

void	crc16xm_calc(Crc16work * crcwork, int n, const void * p)
{
	const BYTE	*c = (const BYTE *)p;

	while ( --n >= 0 )
		crcwork->r = ((crcwork->r) << CHAR_BIT) ^ crc16xm_table[(BYTE)((crcwork->r) >> (16 - CHAR_BIT)) ^ *c++];
	return;
}

crc16_t	crc16xm_getcrc(Crc16work * crcwork)
{
	return ~(crcwork->r) & 0xFFFFu;
}

void		crc16xm_term(Crc16work * crcwork)
{
	return;
}

