/*************************************************************************
*	"crc32.c"
*-------------------------------------------------------------------------
*	$Id: crc32.c 94 2006-09-20 08:51:21Z taro $
*************************************************************************/

#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif
#include	"crc32.h"
#include	<stdlib.h>
#include	<limits.h>

#define	CRCPOLY	(0x04C11DB7ul)	// x^{32}+x^{26}+x^{23}+x^{22}+x^{16}+x^{12}+x^{11]+x^{10}+x^8+x^7+x^5+x^4+x^2+x^1+1

static int		crc32_table_init = 0;
static crc32_t	crc32_table[UCHAR_MAX + 1];
static void 	crc32_maketable(void);

#ifndef	BYTE
#	define	BYTE	unsigned char
#endif

static void	crc32_maketable(void)
{
	unsigned int	i, j;
	crc32_t			r;

	for (i = 0; i <= UCHAR_MAX; i++)
	{
		r = (unsigned long)i << (32 - CHAR_BIT);
		for (j = 0; j < CHAR_BIT; j++)
		{
			if (r & 0x80000000ul)
				r = (r << 1) ^ CRCPOLY;
			else
				r <<= 1;
		}
		crc32_table[i] = r & 0xFFFFFFFFul;
	}
	return;
}

void	crc32_init(Crc32work * crcwork)
{
	if ( crcwork == NULL )
		return;
	if ( crc32_table_init == 0 )
	{
		crc32_maketable();
		crc32_table_init = 1;
	}
	crcwork->r = 0xFFFFFFFFul;
	return;
}

void	crc32_calcbyte(Crc32work * crcwork, unsigned c)
{
	crcwork->r = (crcwork->r << CHAR_BIT) ^ crc32_table[(BYTE)(crcwork->r >> (32 - CHAR_BIT)) ^ c];
	return;
}

void	crc32_calc(Crc32work * crcwork, unsigned size, const void * data)
{
	const BYTE	*c = (const BYTE *)data;

	while ( size-- > 0 )
		crcwork->r = (crcwork->r << CHAR_BIT) ^ crc32_table[(BYTE)(crcwork->r >> (32 - CHAR_BIT)) ^ *c++];
	return;
}

crc32_t		crc32_getcrc(Crc32work * crcwork)
{
	return ~crcwork->r & 0xFFFFFFFFul;
}

void	crc32_ternm(Crc32work * crcwork)
{
	return;
}

