//////////////////////////////////////////////////////////////////////////
//	"crc32.cpp"
//------------------------------------------------------------------------
//	$Id: crc32.cpp 94 2006-09-20 08:51:21Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif
#include	"crc32.h"
#include	<limits.h>

#define	CRCPOLY	(0x04C11DB7UL)	// x^{32}+x^{26}+x^{23}+x^{22}+x^{16}+x^{12}+x^{11]+x^{10}+x^8+x^7+x^5+x^4+x^2+x^1+1

static crc32_t	crctable[UCHAR_MAX + 1];
static void 	makecrctable(void);

Crc32::Crc32()
{
	static int	maketable = 0;

	if ( maketable == 0 )
	{
		::makecrctable();
		maketable = 1;
	}
	this->Init();
}

static void	makecrctable(void)
{
	unsigned int	i, j;
	crc32_t			r;

	for (i = 0; i <= UCHAR_MAX; i++)
	{
		r = (unsigned long)i << (32 - CHAR_BIT);
		for (j = 0; j < CHAR_BIT; j++)
		{
			if (r & 0x80000000ul)
				r = (r << 1) ^ CRCPOLY;
			else
				r <<= 1;
		}
		crctable[i] = r & 0xFFFFFFFFul;
	}
}

void	Crc32::Init(void)
{
	this->r = 0xFFFFFFFFul;
}

void	Crc32::Calc(unsigned c)
{
	this->r = (this->r << CHAR_BIT) ^ ::crctable[(BYTE)(this->r >> (32 - CHAR_BIT)) ^ c];
}

void	Crc32::Calc(unsigned size, const void * data)
{
	BYTE	   *c = (BYTE *)data;

	while ( size-- > 0 )
		r = (r << CHAR_BIT) ^ ::crctable[(BYTE)(r >> (32 - CHAR_BIT)) ^ *c++];
	return;
}

crc32_t		Crc32::GetCrc(void)
{
	return ~this->r & 0xFFFFFFFFul;
}


int		Crc32::CalcFile(HANDLE hFile)
{
	LARGE_INTEGER	size;
	LARGE_INTEGER	fpos;
	BYTE		   *buffer;
	unsigned		bufsiz;

	{
		DWORD	lo, hi;

		if ( (lo = ::GetFileSize(hFile, &hi)) == (DWORD)-1 )
			return -1;
		size.LowPart  = lo;
		size.HighPart = hi;
	}
	if ( size.QuadPart == 0 )
		return 0;
	bufsiz = size.QuadPart < CRC32_MAX_BUFFER_SIZE  ? (unsigned)size.QuadPart : CRC32_MAX_BUFFER_SIZE;
	if ( (buffer = new BYTE [bufsiz]) == NULL )
		return -1;
	fpos.QuadPart = 0;
	::SetFilePointer(hFile, fpos.LowPart, &fpos.HighPart, FILE_BEGIN);
	while ( fpos.QuadPart < size.QuadPart )
	{
		DWORD	dw;

		if ( ::ReadFile(hFile, buffer, bufsiz, &dw, NULL) == 0 )
			return -1;
		if ( dw == 0 )
			break;
		this->Calc(dw, buffer);
		fpos.QuadPart += dw;
	}
	delete [] buffer;
	return 0;
}

