//////////////////////////////////////////////////////////////////////////
//	"md4r_sock.cpp"
//------------------------------------------------------------------------
//	$Id: md4r_sock.cpp 94 2006-09-20 08:51:21Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif
#if	! defined(LINUX)
#if	defined(WIN32)
# include	<winsock2.h>
#endif
#include	"md4r_sock.h"
#if defined(WIN32)
# include	<tchar.h>
# include	<crtdbg.h>
#endif

#if	defined(WIN32) && defined(_DEBUG)
# include	<stdio.h>
# define	DBGPRINT(param_)	printf param_
#else
# define	DBGPRINT(param_)
#endif

const BYTE	MD4R_Sock::md4r_net_id[MD4R_NET_ID_LENGTH         ] = MD4R_NET_ID    ;
const BYTE	MD4R_Sock::md4r_version_id[MD4R_VERSION_ID_LENGTH ] = MD4R_VERSION_ID;
const BYTE	MD4R_Sock::md4r_marker_scdb[MD4R_NET_MARKER_LENGTH] = MD4R_NET_MARKER_SCDB;
const BYTE	MD4R_Sock::md4r_marker_end [MD4R_NET_MARKER_LENGTH] = MD4R_NET_MARKER_END;

MD4R_Sock::MD4R_Sock()
{
	this->recv_timeout = 5000;
	this->send_timeout = 5000;

	this->devwork.hostname = NULL;
	this->devwork.port = 0;

	return;
}

MD4R_Sock::~MD4R_Sock()
{
	if ( this->devwork.hostname != NULL )
		delete [] this->devwork.hostname;

	return;
}

int		MD4R_Sock::SetHostName(const char * hostname)
{
	size_t	len;

	if ( this->devwork.hostname != NULL )
		delete [] this->devwork.hostname;
	len = strlen(hostname);
	if ( (this->devwork.hostname = new char [len+1]) == NULL )
		return -1;
	strcpy(this->devwork.hostname, hostname);
	return 0;
}


int		MD4R_Sock::Read(unsigned size, void * buffer)
{
	fd_set				fds_rcv;
	int					ret, total_read_size = 0;

	while ( total_read_size < (int)size )
	{
		struct timeval		tv_rcv;

		tv_rcv.tv_sec  = this->recv_timeout / 1000;
		tv_rcv.tv_usec = (this->recv_timeout % 1000) * 1000;
		FD_ZERO(&fds_rcv);
		FD_SET(this->GetSocket(), &fds_rcv);

		ret = select((int)this->GetSocket()+1, &fds_rcv, NULL, NULL, &tv_rcv);
		if ( ret == SOCKET_ERROR )
			break;
		if ( FD_ISSET(this->GetSocket(), &fds_rcv) )
		{
			ret = this->Recv(size - total_read_size, (BYTE *)buffer + total_read_size, 0);
			if ( ret <= 0 )
				break;
			total_read_size += ret;
		}
	}

	return total_read_size;
}

int		MD4R_Sock::ReadDummy(unsigned size)
{
	BYTE	buffer[4096];
	int		ret, read_size, total_read_size = 0, remain_size = size;

	while ( remain_size > 0 )
	{
		read_size =  remain_size > sizeof(buffer) ? sizeof(buffer) : remain_size;
		ret = this->Read(read_size, buffer);
		if ( ret <= 0 )
			break;
		total_read_size += ret;
		remain_size -= ret;
	}

	return total_read_size;
}

int			MD4R_Sock::ReadHeader(MD4R_NetCmnHead * chead, unsigned size)
{
	int			ret, pos = 0;
	unsigned	length;

	ret = this->Read(sizeof(*chead), chead);
	if ( ret < sizeof(*chead) )
		return ret;
	pos += ret;
	if ( memcmp(chead->id, MD4R_Sock::md4r_net_id, MD4R_NET_ID_LENGTH) != 0  )
	{
		DBGPRINT(("DEBUG: net_id mismatch!\n"));
		return -1;
	}
	if ( memcmp(chead->version, MD4R_Sock::md4r_version_id, MD4R_VERSION_ID_LENGTH) != 0  )
	{
		DBGPRINT(("DEBUG: version mismatch!\n"));
		return -1;
	}
	length = MD4R_GET32(chead->length);
	if ( length < size )
	{
		DBGPRINT(("DEBUG: Too small header length!\n"));
		return -1;
	}
	if ( size > sizeof(*chead) )
	{
#if	defined(_DEBUG)
		BYTE	*p1, *p2;

		p1 = (BYTE *)chead;
		p2 = (BYTE *)(chead + 1);
#endif
		ret = this->Read( size - sizeof(*chead), chead + 1);
		if ( ret < (int)(size - sizeof(*chead)) )
			return -1;
		pos += ret;
	}

	return pos;
}

int		MD4R_Sock::Write(unsigned size, const void * buffer)
{
	const BYTE			*p = (BYTE *)buffer;
	fd_set				fds_snd;
	int					ret, total_write_size = 0;

	while ( total_write_size < (int)size )
	{
		struct timeval		tv_snd;

		tv_snd.tv_sec  = this->send_timeout / 1000;
		tv_snd.tv_usec = (this->send_timeout % 1000) * 1000;
		FD_ZERO(&fds_snd);
		FD_SET(this->GetSocket(), &fds_snd);

		ret = select((int)this->GetSocket()+1, NULL, &fds_snd, NULL, &tv_snd);
		if ( ret == SOCKET_ERROR )
			break;
		if ( FD_ISSET(this->GetSocket(), &fds_snd) )
		{
			ret = this->Send(size - total_write_size, p, 0);
			if ( ret <= 0 )
				break;
			total_write_size += ret;
			p += ret;
		}
	}

	return total_write_size;
}


//	type 1	[net:]<host>:<port>-<n>-<n>[-<n>]
int		MD4R_Sock::GetNetDrivePath(const TCHAR ** p, unsigned hostsize, TCHAR * host, int * port, int * pathId, int * targetId, int * lun) // static
{
	struct	ArgPar
	{
		const TCHAR	*text;
		unsigned	len;
		int			sep;
	};
	const int	ARG_MAX = 5;
	ArgPar		args[ARG_MAX];
	int			n;

	if ( p == NULL )
		return -1;

	if ( port != NULL )
		*port = 0;
	if ( pathId != NULL )
		*pathId = 0;
	if ( targetId != NULL )
		*targetId = 0;
	if ( lun != NULL )
		*lun = 0;

	if ( _tcsnicmp(*p, TEXT("net"), 3) == 0 )
	{
		 if ( (*p)[3] == TEXT(':') )
		{
			(*p) += 4;
		}
	}
	for ( n = 0; n < ARG_MAX; n++ )
	{
		args[n].text = NULL;
		args[n].len  = 0;
		args[n].sep  = 0;
	}
	n = 0;
	while ( n < ARG_MAX && **p != TEXT('\0') )
	{
		args[n].text = *p;
		if ( n == 0 )	// host
		{
			while ( **p != TEXT('\0') && **p != TEXT(':') )
			{
				args[n].len++;
				(*p)++;
			}
		} else
		{
			while ( **p != '\0' && **p >= TEXT('0') && **p <= TEXT('9') )
			{
				args[n].len++;
				(*p)++;
			}
		}
		args[n].sep = **p;
		n++;
		if ( **p == '\0' || (**p != TEXT(':') && **p != TEXT('-')) )
			break;
		(*p)++;
	}
	if ( n < 4 )
		return -1;

	if ( hostsize > 0 && host != NULL )
	{
		if ( hostsize > args[0].len )
			hostsize = args[0].len;
		else // hostsize <= args[0].len
			hostsize--;
		_tcsncpy(host, args[0].text, hostsize);
		host[hostsize] = TEXT('\0');
	}
	if ( args[0].sep != TEXT(':') )
		return -1;

	if ( port != NULL )
		*port = _ttoi(args[1].text);

	if ( pathId != NULL )
		*pathId = _ttoi(args[2].text);

	if ( targetId != NULL )
		*targetId = _ttoi(args[3].text);

	if ( lun != NULL && n >= 4 )
		*lun = _ttoi(args[4].text);

	return 0;
}

#endif
