//////////////////////////////////////////////////////////////////////////
//	"sqstr.cpp" -- string class
//------------------------------------------------------------------------
//	$Id: sqstr.cpp 94 2006-09-20 08:51:21Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif
#include	"sqstr.h"
#include	<stdlib.h>
#include	<string.h>
#if	defined(WIN32)
# include	<windows.h>
#endif

SQStr::SQStr()
{
	this->p = NULL;
	this->error = SQERR_NOERR;
}

SQStr::SQStr(const SQStr & str)
{
	if ( str.p != NULL )
		((SQStr &)str).p = new srep;
	++(str.p->ref);
	this->p = str.p;
	this->error = this->p != NULL ? SQERR_NOERR : SQERR_OUT_OF_MEMORY;
}

SQStr::SQStr(const TCHAR * s)
{
	this->p = new srep;
	if ( this->p == NULL )
	{
		this->error = SQERR_OUT_OF_MEMORY;
		return;
	}
	this->SetText(s);
}

SQStr::~SQStr()
{
	if ( this->p != NULL )
	{
		if ( --(this->p->ref) == 0 )
		{
			this->ClearText();
			delete this->p;
			this->p = NULL;
		}
	}
}

SQStr &		SQStr::operator = (const SQStr & str)
{
	if ( str.p == NULL )
		((SQStr &)str).p = new srep;
	++(str.p->ref);
	if ( this->p != NULL )
	{
		if ( --(this->p->ref) == 0 )
		{
			this->ClearText();
			delete this->p;
			this->p = NULL;
		}
	}
	this->p = str.p;
	return *this;
}

SQStr &		SQStr::operator = (const TCHAR * s)
{
	if ( this->p != NULL )
	{
		if ( this->p->ref > 1 )
		{
			--(this->p->ref);
			this->p = new srep;
			if ( this->p == NULL )
			{
				this->error = SQERR_OUT_OF_MEMORY;
				return *this;
			}
		} else
			this->ClearText();
	}
	this->SetText(s);
	return *this;
}

void	SQStr::ClearText(void)
{
	if ( this->p != NULL )
	{
		if ( this->p->buffer != NULL )
		{
			if ( (this->p->mode & MODE_FLAG_ALLOC) != 0 )
				delete [] this->p->buffer;
			this->p->buffer = NULL;
		}
		this->p->length = 0;
	}
}

TCHAR *	SQStr::SetText(const TCHAR * s)
{
	this->ClearText();
	if ( s != NULL )
	{
		if ( this->p == NULL )
			this->p = new srep;
		this->p->length = (unsigned)::_tcslen(s);
		if ( (this->p->buffer = new TCHAR[this->p->length+1]) == NULL )
		{
			this->p->length = 0;
			this->error = SQERR_OUT_OF_MEMORY;
			return NULL;	// out of memory
		}
		this->p->mode |= MODE_FLAG_ALLOC;
		::_tcscpy( (TCHAR *)this->p->buffer, s );
	}
	this->error = SQERR_NOERR;
	return this->p != NULL ? (TCHAR *)(this->p->buffer) : NULL;
}

SQStr &	SQStr::SetText(const SQStr & s)
{
	unsigned	size;

	if ( this->p == s.p )
		return *this;
	if ( s.p != NULL )
	{
		switch ( s.p->chcode )
		{
			case SQCHCODE_UNICODE:
				size = (s.p->length + 1) * sizeof(wchar_t);
				break;
			case SQCHCODE_ANSI:
				size = (s.p->length + 1) * sizeof(char);
				break;
			default:
				this->ClearText();
				this->error = SQERR_CHCODE_MISMATCH;
				return *this;
		}
		if ( this->TransferToBuffer(size, s.p->buffer, s.p->length, s.p->chcode) < 0 )
			this->ClearText();
	} else
		this->ClearText();
	return *this;
}

unsigned	SQStr::GetBufferSize(void)
{
	if ( this->p == NULL || this->p->buffer == NULL )	// BugFix 2002.05.31
		return 0;
	switch ( this->p->chcode )
	{
		case SQCHCODE_UNICODE:
			return (this->p->length + 1) * sizeof(wchar_t);
		case SQCHCODE_ANSI:
			return (this->p->length + 1) * sizeof(char);
	}
	return 0;
}

void *	SQStr::SetBuffer(const void * buffer, unsigned length, int chcode)
{
	if ( this->p == NULL )
	{
		if ( (this->p = new srep) == NULL )
			return NULL;
	} else
		this->ClearText();
	this->p->buffer = (void *)buffer;
	this->p->mode  &= ~MODE_FLAG_ALLOC;
	this->p->length = length;
	this->p->chcode = chcode;
	return this->p->buffer;
}

void *	SQStr::CreateBuffer(unsigned size, unsigned length, int chcode)
{
	if ( this->p != NULL )
		this->ClearText();
	else
		this->p = new srep;
	if ( (this->p->buffer = new BYTE [size]) == NULL )
	{
		this->error = SQERR_OUT_OF_MEMORY;
		return NULL;
	}
	this->p->mode |= MODE_FLAG_ALLOC;
	this->p->length = length;
	this->p->chcode = chcode;
	return this->p->buffer;
}

int		SQStr::TransferToBuffer(unsigned byte, const void * buffer, int length, int chcode)
{
	if ( this->p != NULL )
		this->ClearText();
	else
		this->p = new srep;
	if ( buffer != NULL )
	{
		this->p->length = length;
		if ( (this->p->buffer = new BYTE [byte]) == NULL )
		{
			this->p->length = 0;
			this->error = SQERR_OUT_OF_MEMORY;
			return -1;	// out of memory
		}
		this->p->mode |= MODE_FLAG_ALLOC;
		::memcpy( p->buffer, buffer, byte);
		this->p->chcode = chcode;
	}
	this->error = SQERR_NOERR;
	return 0;
}

