//////////////////////////////////////////////////////////////////////////
//	"threadlib.cpp"
//------------------------------------------------------------------------
//	$Id: threadlib.cpp 94 2006-09-20 08:51:21Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif

#include	"threadlib.h"
#if	defined(WIN32)
# include	<process.h>
#elif	defined(LINUX)
# include	<pthread.h>
#endif

//////////////////////////////////////////////////////////////////////////
//	Mutex

MutexLib::MutexLib()
{
	this->created = FALSE;
#if	defined(WIN32)
	this->handle = INVALID_HANDLE_VALUE;
#else
	{
		pthread_mutex_t	mutex_init = PTHREAD_MUTEX_INITIALIZER;

		this->mutex = mutex_init;
	}
#endif
	return;
}

MutexLib::~MutexLib()
{
#if	defined(WIN32)
#else
#endif
	return;
}

#if 0
int		MutexLib::IsCreate(void)
{
	return this->created;
#if	defined(WIN32)
	return this->handle == INVALID_HANDLE_VALUE || this->handle == NULL ? FALSE : TRUE;
#else
	return this->mutex == PTHREAD_MUTEX_INITIALIZER ? FALSE : TRUE;
#endif
}
#endif

int		MutexLib::Create(void)
{
	if ( this->IsCreate() )
		return -1;
#if	defined(WIN32)
	this->handle = CreateMutex(NULL, FALSE, NULL);
	if ( this->handle == NULL )
		return -1;
#else
	pthread_mutex_init(&this->mutex, NULL);
#endif
	this->created = TRUE;
	return 0;
}

int		MutexLib::Destroy(void)
{
	if ( this->IsCreate() == FALSE )
		return -1;
#if	defined(WIN32)
	if ( this->handle != INVALID_HANDLE_VALUE && this->handle != NULL )
		::CloseHandle(this->handle);
	this->handle = INVALID_HANDLE_VALUE;
#else
	{
		pthread_mutex_t	mutex_init = PTHREAD_MUTEX_INITIALIZER;

		pthread_mutex_destroy(&this->mutex);
		this->mutex = mutex_init;
	}
#endif
	this->created = FALSE;

	return 0;
}

int		MutexLib::Lock(void)
{
	if ( this->IsCreate() == FALSE )
		return -1;
#if	defined(WIN32)
	{
		int		res;

		res = ::WaitForSingleObject(this->handle, INFINITE);
	}
#else
	pthread_mutex_lock(&this->mutex);
#endif
	return 0;
}

int		MutexLib::Unlock(void)
{
	if ( this->IsCreate() == FALSE )
		return -1;
#if	defined(WIN32)
	{
		int		res;

		res = ::ReleaseMutex(this->handle);
	}
#else
	pthread_mutex_unlock(&this->mutex);
#endif
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Thread

ThreadLib::ThreadLib()
{
	this->runFlag = FALSE;
	this->arg     = NULL;
	this->func    = (ThreadFunc)0;
#if	defined(WIN32)
	this->threadHandle = NULL;
	this->threadId     = 0;
#else
#endif

	return;
}

ThreadLib::~ThreadLib()
{
	this->Destroy();
	this->runMutex.Destroy();
	return;
}

int		ThreadLib::Create(ThreadLib::ThreadFunc func, void * arg)
{
	int		retcode;

	if ( func == NULL || this->IsRunning() )
		return -1;
	if ( this->runMutex.IsCreate() == FALSE )
		this->runMutex.Create();
	this->func = func;
	this->arg  = arg;
	this->runMutex.Lock();
#if	defined(WIN32)
	if ( (this->threadHandle = (HANDLE) _beginthreadex( NULL, 0, start_thread, this, 0, &this->threadId)) == 0 )
	{	// error
		this->runFlag = FALSE;
		retcode = -1;
	} else
	{
		this->runFlag = TRUE;
		retcode = 0;
	}
#else
	if ( pthread_create(&this->thread, NULL, start_thread, this) != 0 )
	{	// ERROR
		this->runFlag = FALSE;
		retcode = -1;
	} else
	{
		this->runFlag = TRUE;
		retcode = 0;
	}
#endif
	this->runMutex.Unlock();

	return retcode;
}

int		ThreadLib::Destroy(void)
{
	if ( this->IsRunning() )
		return -1;
#if	defined(WIN32)
	if ( this->threadHandle != NULL )
	{
		::CloseHandle(this->threadHandle);
		this->threadHandle = NULL;
	}
#else
#endif
	return 0;
}

int		ThreadLib::WaitExit(void)
{
	if ( this->IsRunning() == FALSE )
		return -1;
#if	defined(WIN32)
	{
		int		res;

		res = ::WaitForSingleObject(this->threadHandle, INFINITE);
	}
#else
	pthread_join(this->thread, NULL);
#endif
	return 0;
}

#if	defined(WIN32)
unsigned	__stdcall	ThreadLib::start_thread(void * arg)	// static
{
	ThreadLib	*thread = (ThreadLib *) arg;

	if ( thread->func )
		(thread->func)(thread->arg);
	thread->endthread();
	return 0;
}
#else
void *	ThreadLib::start_thread(void * arg) // static
{
	ThreadLib	*thread = (ThreadLib *) arg;

	if ( thread->func )
		(thread->func)(thread->arg);
	thread->endthread();
	return NULL;
}
#endif

void	ThreadLib::endthread(void)
{
	if ( this->IsRunning() == FALSE )
		return;
	this->runMutex.Lock();
	this->runFlag = FALSE;
	this->runMutex.Unlock();
#if	defined(WIN32)
	_endthreadex(0);
#else
#endif
	return;
}

