//////////////////////////////////////////////////////////////////////////
//	"dev_storagegethotpluginf.cpp"
//------------------------------------------------------------------------
//	$Id: dev_fileio.cpp 94 2006-09-20 08:51:21Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"devctl.h"

#if defined(DEVCTL_USE_FILEIO)
int		DevCtl::CreateCtlWork(void)
{
	DISK_GEOMETRY	geo;
	int				devtype = -1;

	if ( this->handle == NULL )
		return -1;
	if ( this->ctl != NULL )
		return -1;
	if ( this->DiskGetDriveGeometry(geo) == 0 )
	{
	} else if ( this->StorageGetMediaTypes(geo) == 0 )
	{
	} else if ( this->CdromGetDriveGeometry(geo) == 0 )
	{
		devtype = DEV_CDROM;
	} else
		return -1;

	if ( (this->ctl = new CtlWork) == NULL )
		return -1;
	this->ctl->geo = geo;
	this->ctl->maxBlk = geo.Cylinders.QuadPart * (INT64)geo.TracksPerCylinder * (INT64)geo.SectorsPerTrack;
	this->ctl->sectorSize = geo.BytesPerSector;
	if ( devtype < 0 )
	{
		switch ( geo.MediaType )
		{
			case F5_1Pt2_512:
			case F5_360_512:
			case F5_320_512:
			case F5_320_1024:
			case F5_180_512:
			case F5_160_512:
			case F5_640_512:
			case F5_720_512:
			case F5_1Pt23_1024:
			case F3_1Pt44_512:
			case F3_2Pt88_512:
			case F3_20Pt8_512:
			case F3_720_512:
			case F3_120M_512:
			case F3_640_512:
			case F3_1Pt2_512:
			case F3_1Pt23_1024:
			case F3_128Mb_512:
			case F3_230Mb_512:
			case F3_200Mb_512:
			case F3_240M_512:
			case F3_32M_512:
			case F8_256_128:
				devtype = DEV_FLOPPY;
				break;

			case RemovableMedia:
				devtype = DEV_REMOVABLE;
				break;

			case FixedMedia:
				devtype = DEV_FIXED;
				break;

			case Unknown:
			default:
				devtype = DEV_UNKNOWN;
				break;
		}
	}
	this->ctl->devtype = devtype;

	return 0;
}

int		DevCtl::ReadBlk(INT64 pos, unsigned count, void * buffer)
{
	DWORD	dwResult;

	if ( this->handle == NULL || this->ctl == NULL )
		return -1;
	if ( pos < 0 || pos >= this->ctl->maxBlk )
		return -1;
	if ( this->SetFilePointer(pos, FPOS_BEGIN) < 0 )
		return -1;
	if ( ::ReadFile(this->handle, buffer, this->ctl->sectorSize * count, &dwResult, NULL) == FALSE )
		return -1;
	return (int) dwResult / this->ctl->sectorSize;
}

int		DevCtl::WriteBlk(INT64 pos, unsigned count, const void * buffer)
{
	DWORD	dwResult;

	if ( this->handle == NULL || this->ctl == NULL )
		return -1;
	if ( pos < 0 || pos >= this->ctl->maxBlk )
		return -1;
	if ( this->SetFilePointer(pos, FPOS_BEGIN) < 0 )
		return -1;
	if ( ::WriteFile(this->handle, buffer, this->ctl->sectorSize * count, &dwResult, NULL) == FALSE )
		return -1;
	return (int) dwResult / this->ctl->sectorSize;
}


int		DevCtl::GetDiskFormat(void)
{
	BYTE		   *buffer = NULL;
	int				retcode;

	if ( this->handle == NULL || this->ctl == NULL )
		return -1;
	try
	{
		if ( (buffer = new BYTE [this->ctl->sectorSize]) == NULL )
			throw -1;
		if ( this->ReadBlk( 0I64, 1, buffer) < 1 )
			throw -1;

		if ( buffer[this->ctl->sectorSize - 2] == 0x55 && buffer[this->ctl->sectorSize-1] == 0xAA )	// IBM
		{
			if ( (buffer[0] == 0xE9 || buffer[0] == 0xEB)
			  &&  *(WORD *)(buffer + 0x0B) == this->ctl->sectorSize 
			  && (buffer[0x015] == 0xF0 || buffer[0x015] == 0xF8) )
			{
				throw (int)DF_IBM_FD;
			} else
			{
				throw (int)DF_IBM_HD;
			}
		} else if ( buffer[0] == 'I' && buffer[1] == 'P' && buffer[3] == 'L' )	// FMR HDD
		{
			throw (int)DF_FMR_HD;
		} else
		{
			throw (int)DF_UNKNOWN;
		}
	}
	catch (int code) { retcode = code; }
	if ( buffer != NULL )
		delete [] buffer;
	return retcode;
}
#endif
