//////////////////////////////////////////////////////////////////////////
//	"devctl.h" for Linux/Win32
//------------------------------------------------------------------------
//	$Id: devctl.h 110 2006-09-21 18:14:40Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	DEVCTL_H_INCLUDED
#define	DEVCTL_H_INCLUDED

#if	defined(WIN32)
# include	<windows.h>
# include	<winioctl.h>
# include	<ntddcdrm.h>
# include	<ntddscsi.h>
#elif	defined(LINUX)
# include	<stdlib.h>
# include	<stdio.h>
# include	<string.h>
# include	<errno.h>
# include	<unistd.h>
# include	<fcntl.h>
# include	<sys/ioctl.h>
# include	<sys/stat.h>
# include	<scsi/sg.h>
# include	"../linux/defs.h"
# if	!defined(ATTB_PACKED)
#  define	ATTB_PACKED	__attribute ((packed))
# else
#  define	ATTB_PACKED
# endif
#endif

//#define	DEVCTL_USE_FILEIO

#if	defined(DEVCTL_USE_MD4RNET)
class	MD4R_Sock;
typedef struct MD4R_NetScdb_tag MD4R_NetScdb;
#endif

class	DevCtl
{
public:
#if	defined(WIN32)
	typedef __int64 	INT64;
#elif	defined(LINUX)
	typedef long long	INT64;
#endif

	enum rwmode_enum
	{
		RW_READ        = 0x01,
		RW_WRITE       = 0x02,
		RW_SYNC        = 0x04,
		RW_NONBLOCK    = 0x08,
		RW_SHARE_READ  = 0x10,
		RW_SHARE_WRITE = 0x20,
		RW_ASYNC       = 0x40,	// linux no implementation
		RW_NOBUFFER    = 0x80,	// linux no implementation
	};
	enum handle_type_enum
	{
		HT_UNKNOWN,
#if	defined(WIN32)
		HT_DISK ,
		HT_DRIVE,
		HT_SCSI ,
#elif	defined(LINUX)
		HT_BLKDEV ,
		HT_CDROM  ,
		HT_FILE   ,
		HT_SG     ,
#endif
#if	defined(DEVCTL_USE_MD4RNET) && defined(WIN32)
		HT_NET  ,
#endif
	};
#if	defined(WIN32)
	enum filepos_method_enum
	{
		FPOS_BEGIN   = FILE_BEGIN,
		FPOS_END     = FILE_END,
		FPOS_CURRENT = FILE_CURRENT,
	};
#endif
#if defined(DEVCTL_USE_FILEIO) && defined(WIN32)
	enum devtype_enum
	{
		DEV_UNKNOWN  ,
		DEV_FLOPPY   ,
		DEV_FIXED    ,
		DEV_REMOVABLE,
		DEV_CDROM    ,
	};
	enum diskformat_enum
	{
		DF_UNKNOWN,
		DF_IBM_FD,
		DF_IBM_HD,
		DF_FMR_HD,
	};
	struct CtlWork
	{
		int				devtype;
		DISK_GEOMETRY	geo;
		INT64			maxBlk;
		unsigned		sectorSize;
	};
#endif

#if	defined(LINUX)
	//////////////////////////////////////////////////////////////////////
	//	Block Device
private:
	struct tag_blk
	{
		unsigned		blkLen;	// sector size
		INT64			blkNum;
	} blk;
#endif

	//////////////////////////////////////////////////////////////////////
	// SCSI
public:
#if	defined(SCSIHEAD_H_INCLUDED)
# undef	SCSIHEAD_H_INCLUDED
#endif
#	include	"scsihead.h"
#if	defined(WIN32)
	class ScsiData;
	class ScsiBusData;
	class ScsiInqData;

	class ScsiDevData
	{
	friend class	DevCtl;
	private:
		unsigned		numBus;
		ScsiBusData	   *busData;
	public:
		ScsiDevData();
		~ScsiDevData();
		unsigned		GetBusNum(void) const { return this->numBus; }
		ScsiBusData	*	GetBusData(unsigned n) const { return this->busData == NULL || n >= this->numBus ? NULL : &this->busData[n]; }
	};
	class ScsiBusData
	{
	friend class	DevCtl;
	private:
		unsigned		numUnits;
		unsigned		busId;
		ScsiInqData	   *inqData;
	public:
		ScsiBusData();
		~ScsiBusData();
		unsigned		GetInqNum(void) const { return this->numUnits; }
		unsigned		GetBusId(void) const { return this->busId; }
		ScsiInqData	*	GetInqData(unsigned n) const { return this->inqData == NULL || n >= this->numUnits ? NULL : &this->inqData[n]; }
	};
#endif

	class ScsiInqData
	{
	friend class	DevCtl;
	private:
		unsigned		pathId;
		unsigned		targetId;
		unsigned		lun;
		BOOL			devClaimed;
		unsigned		size;
		BYTE *			data;	// SCSIINQUIRY *
	public:
		BOOL			unknownId;
	public:
		ScsiInqData();
		~ScsiInqData();
		unsigned		GetPathId(void) const { return this->pathId; }
		unsigned		GetTargetId(void) const { return this->targetId; }
		unsigned		GetLun(void) const { return this->lun; }
		BOOL			GetClaimed(void) const { return this->devClaimed; }
		unsigned		GetDataSize(void) const { return this->size; }
		BYTE *			GetData(void) const { return this->data; }
		SCSIINQUIRY *	GetInquiryData(void) const { return (SCSIINQUIRY *)this->data; }
		int				Copy(ScsiInqData & src);
	};

	enum scsi_enum
	{
		SCSI_SENSE_LEN       = 14,	// Default sense buffer length
		SCSI_CDB6_LEN        = 6,
		SCSI_CDB10_LEN       = 10,
		SCSI_CDB12_LEN       = 12,
		SCSI_CDB16_LEN       = 16,
		SCSI_CDBMAX_LEN      = 16,
		SCSI_TIMEOUT         = 5000, // 10,
		SCSI_TIMEOUT_SEC     = 30,	// for Linux
		SCSI_SPT_BUFFER_SIZE = 256,
	};
	enum scsi_status_enum
	{
		SCSI_STATUS_GOOD                       = 0x00,
		SCSI_STATUS_CHECK_CONDITION            = 0x02,
		SCSI_STATUS_CONDITION_MET              = 0x04,
		SCSI_STATUS_BUSY                       = 0x08,
		SCSI_STATUS_INTERMEDIATE               = 0x10,
		SCSI_STATUS_INTERMEDIATE_CONDITION_MET = 0x14,
		SCSI_STATUS_RESERVATION_CONFLICT       = 0x18,
		SCSI_STATUS_COMMAND_TERMINATED         = 0x22,
		SCSI_STATUS_QUEUE_FULL                 = 0x28,
	};
	enum	ScsiCapacityMode
	{
		SCSI_CAPACITYMODE_COUNT ,
		SCSI_CAPACITYMODE_MAXLBA,
	};
#if	defined(WIN32)
public:
#if	defined(_MSC_VER)
#	pragma pack(push,1)
#endif
	struct SPT_BUFFER
	{
		SCSI_PASS_THROUGH	spt;
		ULONG				filler;      // realign buffers to double word boundary
		BYTE				senseArea[SCSI_SENSE_LEN+2];
		UCHAR				buffer[SCSI_SPT_BUFFER_SIZE];
	};
	struct SPT_DIRECT_BUFFER
	{
		SCSI_PASS_THROUGH_DIRECT	spt;
		ULONG						filler;      // realign buffers to double word boundary
		BYTE						senseArea[SCSI_SENSE_LEN+2];
	};
#if	defined(_MSC_VER)
#	pragma pack(pop)
#endif

#endif
#if	defined(LINUX)
public:
	struct Sgt
	{
		int				fd;
		unsigned		targetId;
		unsigned		targetLun;
		sg_io_hdr_t		hdr;
		unsigned char	cdb[SCSI_CDBMAX_LEN];
		unsigned char	sense[SCSI_SENSE_LEN];
	};
#endif

public:
	static unsigned	ScsiUint32(const BYTE p[4]) { return (p[0]<<24)+(p[1]<<16)+(p[2]<<8)+p[3]; }
	static unsigned	ScsiUint24(const BYTE p[3]) { return (p[0]<<16)+(p[1]<<8)+p[2]; }
	static unsigned	ScsiUint16(const BYTE p[2]) { return (p[0]<<8)+p[1]; }

private:
	TCHAR		   *devname;
#if	defined(WIN32)
	HANDLE			handle;
#elif	defined(LINUX)
	int				fd;
#endif
	unsigned		htype;		// handle type
	unsigned		rwmode;
#if defined(DEVCTL_USE_FILEIO) && defined(WIN32)
	CtlWork			*ctl;
#endif
	unsigned		scsi_timeout;
#if	defined(WIN32)
	HANDLE			hEvent;
	OVERLAPPED		ol;
#endif
	unsigned		maxCdbSize;
#if	defined(LINUX)
	struct stat	statbuf;
#endif
	static ScsiCapacityMode	scsiCapacityMode;

public:
	DevCtl();
	~DevCtl();

	int			OpenDevice(const TCHAR * name, unsigned rwmode = RW_READ|RW_WRITE|RW_SHARE_READ);
	int			CloseDevice(void);

#if	defined(WIN32)
	HANDLE			GetHandle(void) const { return this->handle; }
	unsigned		GetHandleType(void) const { return this->htype; }
	int				GetFileType(void);
	int				GetFileSize(LARGE_INTEGER & size);
	INT64			GetFileSize(void);
	int				SetFilePointer(LARGE_INTEGER & pos, int method);
	int				SetFilePointer(INT64 & pos, int method);

	int				DeviceIoControl(DWORD ioctl, void * inbuffer, unsigned insize, void * outbuffer, unsigned outsize, DWORD * byteRet = NULL);
	int				DeviceIoControl(DWORD ioctl, void * inbuffer, unsigned insize, void * outbuffer, unsigned outsize, DWORD * byteRet, OVERLAPPED & overlapped);
	int				DiskIsWritable(void);
	int				DiskGetDriveGeometry(DISK_GEOMETRY & geo);
	int				DiskGetPartitionInfo(PARTITION_INFORMATION & partInfo);
	int				StorageGetMediaTypes(DISK_GEOMETRY & geo);
	int				CdromGetDriveGeometry(DISK_GEOMETRY & geo);
	int				CdromReadToc(unsigned size, CDROM_TOC * toc);
	int				CdromReadToc(DevCtl::SCSITOCDATA & toc);
#if (_WIN32_WINNT >= 0x500)
	int				DiskGetLengthInfo(GET_LENGTH_INFORMATION & lenghInfo);
	int				DiskGetDriveGeometryEx(DISK_GEOMETRY_EX & geo);
#endif
	int				StorageGetHotPlugInfo(STORAGE_HOTPLUG_INFO & hi);

	static const TCHAR *	GetMediaTypeName(int mediatype);
#endif

#if defined(DEVCTL_USE_FILEIO) && defined(WIN32)
	int				CreateCtlWork(void);
	int				ReadBlk(INT64 pos, unsigned count, void * buffer);
	int				WriteBlk(INT64 pos, unsigned count, const void * buffer);
	int				GetSectorSize(void) const { return this->handle == NULL || this->ctl == NULL ? -1 : (int)this->ctl->sectorSize; }
	int				GetMaxBlk(void) const { return this->handle == NULL || this->ctl == NULL ? -1 : (int)this->ctl->maxBlk; }
	int				GetDiskFormat(void);
#endif

#if	defined(LINUX)
	int			GetHandle(void) const { return this->fd; }
	int			GetHType(void) { return this->fd < 0 ? -1 : (int)this->htype; }
#endif

#if	defined(LINUX)
	//////////////////////////////////////////////////////////////////////
	// Block Device
private:
	int			blkdev_opendevice(void);
public:
	unsigned	GetDefaultBlockSize(void) const { return this->blk.blkLen; }
	int			SetDefaultBlockSize(unsigned size);
	BOOL		IsBlkDev(void) { return this->fd >= 0 && S_ISBLK(this->statbuf.st_mode) ? TRUE : FALSE; }
	int			BlkDevGetCapacity(unsigned & blkLen, INT64 & blkNum);
	int			BlkDeviInquiry(DevCtl::SCSIINQUIRY & inq);
private:
	int			cdrom_opendevice(void);
public:
	int			CdromReadToc(BOOL msf, unsigned size, void * buffer);
#endif

	//////////////////////////////////////////////////////////////////////
	// SCSI
public:
	void			SetMaxCdbSize(unsigned size) { this->maxCdbSize = size; }
	int				GetMaxCdbSize(void) const { return this->maxCdbSize; }
	unsigned		GetScsiTimeout(void) const { return this->scsi_timeout; }
	static INT64 ScsiGetLBA(BYTE b[4]) { return ((INT64)b[0]<<24) + ((INT64)b[1]<<16) + ((INT64)b[2]<<8) + ((INT64)b[3]); }
	static void  ScsiSetLBA(BYTE b[4], INT64 size) { b[0] = (BYTE)(size>>24); b[1] = (BYTE)(size>>16); b[2] = (BYTE)(size>>8); b[3] = (BYTE)size; }
	static INT64 ScsiGet32(BYTE b[4]) { return SCSI_GET32(b); }
	static INT64 ScsiGet64(BYTE b[8]) { return SCSI_GET64(b); }
#if	defined(WIN32)
	int				ScsiGetAddress(SCSI_ADDRESS & scsiAddr);
	int				ScsiGetInquiryData(unsigned size, void * buffer);
	ScsiDevData *	ScsiGetInquiryData(void);
#endif
	int				ScsiCreateInqData(DevCtl::ScsiInqData & inqData, int pathId, int targetId, int lun);
#if	defined(WIN32)
	int				ScsiGetCapabilities(IO_SCSI_CAPABILITIES & capa);
	int				ScsiPassThrough(unsigned insize, SCSI_PASS_THROUGH * inspt, unsigned outsize, SCSI_PASS_THROUGH * outspt);
	int				ScsiPassThroughDirect(unsigned insize, SCSI_PASS_THROUGH_DIRECT * inspt, unsigned outsize, SCSI_PASS_THROUGH_DIRECT * outspt);
	int				ScsiPassThroughDirect(unsigned insize, SCSI_PASS_THROUGH_DIRECT * inspt, unsigned outsize, SCSI_PASS_THROUGH_DIRECT * outspt, OVERLAPPED & overlapped);
	int				ScsiCheckCondition(BYTE * senseData);
#endif
#if	defined(LINUX)
	int				ScsiCheckCondition(DevCtl::Sgt & sgt);
#endif
	int				ScsiInquiry(int pathId, int targetId, int lun, int pageCode, unsigned size, void * buffer);
	int				ScsiTestUnitReady(int pathId, int targetId, int lun);
	int				ScsiLogSense(int pathId, int targetId, int lun, BOOL sp, BOOL ppc, unsigned pc, int pageCode, unsigned point, unsigned size, void * buffer);
	int				ScsiModeSense(int pathId, int targetId, int lun, BOOL blockDesc, unsigned pc, int pageCode, unsigned size, void * buffer);
	int				ScsiModeSense(int pathId, int targetId, int lun, int pageCode, unsigned size, void * buffer) {
					return this->ScsiModeSense(pathId, targetId, lun, FALSE, 0, pageCode, size, buffer); }
	int				ScsiModeSelect(int pathId, int targetId, int lun, BOOL pageFormat, BOOL savePages, unsigned size, const void * buffer);
	int				ScsiReadCapacity(int pathId, int targetId, int lun, unsigned size, void * buffer);
	int				ScsiRead(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, void * buffer);
#if	defined(WIN32)
	int				ScsiRead(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, void * buffer, SPT_DIRECT_BUFFER & spt, OVERLAPPED & overlapped);
#endif
	int				ScsiWrite(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, const void * buffer);
#if	defined(WIN32)
	int				ScsiWrite(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, const void * buffer, SPT_DIRECT_BUFFER spt, OVERLAPPED & overlapped);
#endif
	int				ScsiVerify(int pathId, int targetId, int lun, INT64 addr, unsigned count);
	int				ScsiVerify(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, const void * buffer, BOOL check);
	int				ScsiSeek(int pathId, int targetId, int lun, INT64 addr);
	int				ScsiReadToc(int pathId, int targetId, int lun, BOOL msf, unsigned size, void * buffer);
	int				ScsiReadToc(int pathId, int targetId, int lun, BOOL msf, DevCtl::SCSITOCDATA & toc);
	int				ScsiReadDefectData(int pathId, int targetId, int lun, unsigned list, unsigned fmt, unsigned size, void * buffer);
	int				ScsiGetConfiguration(int pathId, int targetId, int lun, unsigned rt, unsigned start, unsigned size, void * buffer);
	int				ScsiReadDvdStructure(int pathId, int targetId, int lun, unsigned addr, unsigned format, unsigned layer, unsigned size, void * buffer);
	int				ScsiSetCdSpeed(int pathId, int targetId, int lun, unsigned rc, unsigned speed);
	int				ScsiStartStopUnit(int pathId, int targetId, int lun, unsigned loadeject_start);
	int				ScsiLoadUnloadMedium(int pathId, int targetId, int lun, unsigned loadunload_start);
	int				ScsiPreventAllowMediumRemoval(int pathId, int targetId, int lun, unsigned state);

	static TCHAR *	ScsiGetDeviceType(unsigned devType, TCHAR * buffer = NULL);

	//////////////////////////////////////////////////////////////////////
	// SMART
public:
#	include	"atahead.h"

#if	defined(WIN32)
	enum	SmartIoCtl
	{
		SMART_IOCTL_RCV  = SMART_RCV_DRIVE_DATA    ,
		SMART_IOCTL_SEND = SMART_SEND_DRIVE_COMMAND,
	};
	int				SmartGetVersion(GETVERSIONINPARAMS & ver);
	int				SmartIoControl(SmartIoCtl ioctl, unsigned inp_size, SENDCMDINPARAMS * inp_buffer, unsigned out_size, SENDCMDOUTPARAMS * out_buffer);
#endif
	int				AtaReadIndentify(int drive, DevCtl::AtaIdentifyDevice & identify);
	int				SmartReadValues(int drive, DevCtl::AtaSmartValues & val);
	int				SmartReadThresholds(int drive, DevCtl::AtaSmartThresholds & thr);

	//////////////////////////////////////////////////////////////////////
	// MD4RNet
private:
#if	defined(DEVCTL_USE_MD4RNET) && defined(WIN32)
	MD4R_Sock	*socklib;

private:
	void		DestroySock(void);
public:
	BOOL		IsNetDevice(void) const { return this->htype == HT_NET && this->socklib != NULL ? TRUE : FALSE; }
	int			NetSet(const TCHAR * host, int port);
	int			NetOpen(unsigned rwmode);
	int			NetClose(void);
	int			NetScsiSend(MD4R_NetScdb * scdb, unsigned outSize, void * outData);
	int			NetScsiRecv(MD4R_NetScdb * scdb, unsigned senseSize, void * senseData, unsigned inpSize, void * inpBuffer);
	int			NetScsiCommand(unsigned insize, SCSI_PASS_THROUGH * inspt, unsigned outsize, SCSI_PASS_THROUGH * outspt);
	int			NetScsiCommand(unsigned insize, SCSI_PASS_THROUGH_DIRECT * inspt, unsigned outsize, SCSI_PASS_THROUGH_DIRECT * outspt);
#endif
};

#endif
