//////////////////////////////////////////////////////////////////////////
//	"perflib.cpp"
//------------------------------------------------------------------------
//	$Id: perflib.cpp 163 2006-02-25 12:25:23Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"perflib.h"

#if	defined(_WIN32)
#pragma	comment(lib, "pdh.lib")
#endif

PerfLib::CounterWork::CounterWork(void)
{
	this->prev = this->next = NULL;
	this->enableCounter = FALSE;
	this->hCounter = NULL;
	this->path = NULL;
}

PerfLib::CounterWork::~CounterWork(void)
{
	if ( this->enableCounter != FALSE )
		::PdhRemoveCounter(this->hCounter);
	this->enableCounter = FALSE;
	this->hCounter = NULL;
	if ( this->path != NULL )
		delete [] this->path;
}

//////////////////////////////////////////////////////////////////////////
//	PerfLib

PerfLib::PerfLib()
{
	this->openQuery = FALSE;
	this->numCounter = 0;
	this->counterList = NULL;
	this->msgbuf = NULL;
}

PerfLib::~PerfLib()
{
	while ( this->numCounter > 0 && this->counterList != NULL )
		this->RemoveCounter(this->counterList);

	if ( this->openQuery )
		this->CloseQuery();

	if ( this->msgbuf != NULL )
		::LocalFree(this->msgbuf);

}

void	PerfLib::add_counter(PerfLib::Counter counter)
{
	counter->next = this->counterList;
	if ( this->counterList != NULL )
		this->counterList->prev = counter;
	this->counterList = counter;
	this->numCounter++;
}

void	PerfLib::del_counter(PerfLib::Counter counter)
{
	if ( counter->next != NULL )
		counter->next->prev = counter->prev;
	if ( counter->prev != NULL )
		counter->prev->next = counter->next;
	else
		this->counterList = counter->next;
	counter->next = counter->prev = NULL;
	this->numCounter--;
}

int		PerfLib::OpenQuery(void)
{
	if ( this->openQuery )
		return 0;
	if ( ::PdhOpenQuery(NULL, 0, &hQuery) != ERROR_SUCCESS )
		return -1;
	this->openQuery = TRUE;
	return 0;
}

void	PerfLib::CloseQuery(void)
{
	if ( this->openQuery == FALSE )
		return;
	::PdhCloseQuery(this->hQuery);
	this->openQuery = FALSE;
	this->hQuery = NULL;
}

PerfLib::Counter	PerfLib::AddCounter(const TCHAR * path, void * userdata)
{
	Counter		counter;

	if ( (counter = new CounterWork) == NULL )
		return NULL;
	if ( (counter->path = new TCHAR [_tcslen(path)+1]) == NULL )
		return NULL;
	_tcscpy(counter->path, path);
	if ( (this->pdhstt = ::PdhAddCounter(this->hQuery, counter->path, (DWORD_PTR)userdata, &counter->hCounter)) != ERROR_SUCCESS )
	{
		delete counter;
		return NULL;
	}
	counter->enableCounter = TRUE;
	this->add_counter(counter);
	return counter;
}

PerfLib::Counter	PerfLib::AddCounter(PDH_COUNTER_PATH_ELEMENTS & ele, void * userdata)
{
	TCHAR	path[PDH_MAX_COUNTER_PATH];
	DWORD	pathSize;

	pathSize = sizeof(path)/sizeof(*path);
	if ( (this->pdhstt = ::PdhMakeCounterPath(&ele, path, &pathSize, 0)) != ERROR_SUCCESS )
		return NULL;
	return this->AddCounter(path, userdata);
}

int				PerfLib::RemoveCounter(PerfLib::Counter counter)
{
	if ( counter == NULL )
		return -1;
	if ( counter->enableCounter )
	{
		::PdhRemoveCounter(counter->hCounter);
		counter->enableCounter = FALSE;
	}
	if ( this->numCounter > 0 && this->counterList != NULL )
		this->del_counter(counter);
	delete counter;
	return 0;
}

int		PerfLib::CalculateCounterFromRawValue(DWORD format, PDH_RAW_COUNTER * rawValue1, PDH_RAW_COUNTER * rawValue2, PDH_FMT_COUNTERVALUE * fmtValue)
{
	if ( this->openQuery == FALSE )
		return -1;
	if ( (this->pdhstt = ::PdhCalculateCounterFromRawValue(this->hQuery, format, rawValue1, rawValue2, fmtValue)) != ERROR_SUCCESS )
		return -1;
	return 0;
}

int		PerfLib::CollectQueryData(void)
{
	if ( this->openQuery == FALSE )
		return -1;
	if ( (this->pdhstt = ::PdhCollectQueryData(this->hQuery)) != ERROR_SUCCESS )
		return -1;
	return 0;
}

int		PerfLib::GetFormattedCounterValue(PerfLib::Counter counter, DWORD format, DWORD * type, PDH_FMT_COUNTERVALUE * value)
{
	if ( counter == NULL || counter->enableCounter == FALSE || this->openQuery == FALSE )
		return -1;
	if ( (this->pdhstt = ::PdhGetFormattedCounterValue(counter->hCounter, format, type, value)) != ERROR_SUCCESS )
		return -1;
	return 0;
}
int		PerfLib::GetRawCounterValue(PerfLib::Counter counter, DWORD * type, PDH_RAW_COUNTER *rawcounter)
{
	if ( counter == NULL || counter->enableCounter == FALSE || this->openQuery == FALSE )
		return -1;
	if ( (this->pdhstt = ::PdhGetRawCounterValue(counter->hCounter, type, rawcounter)) != ERROR_SUCCESS )
		return -1;
	return 0;
}

const TCHAR *	PerfLib::GetErrMsg(DWORD stt)
{
	int		ret;

	if ( this->msgbuf != NULL )
	{
		::LocalFree(this->msgbuf);
		this->msgbuf = NULL;
	}
	ret = ::FormatMessage(
	      FORMAT_MESSAGE_ALLOCATE_BUFFER
	    | FORMAT_MESSAGE_FROM_SYSTEM
	    | FORMAT_MESSAGE_IGNORE_INSERTS
	    , NULL
	    , stt
	    , MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT) // ̌
	    , (LPTSTR) &this->msgbuf
	    , 0
	    , NULL );
	if ( ret == 0 )
	{
		ret = ::FormatMessage(
		      FORMAT_MESSAGE_ALLOCATE_BUFFER
		    | FORMAT_MESSAGE_FROM_HMODULE
		    | FORMAT_MESSAGE_IGNORE_INSERTS
		    , ::GetModuleHandle(TEXT("pdh.dll"))
		    , stt
		    , MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT) // ̌
		    , (LPTSTR) &this->msgbuf
		    , 0
		    , NULL );
	}

	return ret > 0 ? this->msgbuf : TEXT("");
#if 0
	int		stt;

	switch ( stt )
	{
		case PDH_CSTATUS_BAD_COUNTERNAME  : stt = 0; break;
		case PDH_CSTATUS_NO_COUNTER       : stt = 1; break;
		case PDH_CSTATUS_NO_COUNTERNAME   : stt = 2; break;
		case PDH_CSTATUS_NO_MACHINE       : stt = 3; break;
		case PDH_CSTATUS_NO_OBJECT        : stt = 4; break;
		case PDH_FUNCTION_NOT_FOUND       : stt = 5; break;
		case PDH_INVALID_ARGUMENT         : stt = 6; break;
		case PDH_INVALID_HANDLE           : stt = 7; break;
		case PDH_MEMORY_ALLOCATION_FAILURE: stt = 8; break;
		default:
			stt = this->pdhstt;
			break;
	}
#endif
}
